@@ -7,13 +7,15 @@
    Manger and several other people whose names I've misplaced.
 
    Available from http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.c. Last
-   updated 8 June 2013 (version 20130608, if you prefer it that way).  
+   updated 8 August 2015 (version 20150808, if you prefer it that way).  
    To build under Windows, use 'cl /MD dumpasn1.c'.  To build on OS390 or 
    z/OS, use '/bin/c89 -D OS390 -o dumpasn1 dumpasn1.c'.
 
    This code grew slowly over time without much design or planning, and with
    extra features being tacked on as required.  It's not representative of my
-   normal coding style.  cryptlib,
+   normal coding style, and should only be used as a debugging/diagnostic 
+   tool and not in a production environment (I'm not sure how you'd use
+   it in production anyway, but felt I should point that out).  cryptlib,
    http://www.cs.auckland.ac.nz/~pgut001/cryptlib/, does a much better job of
    checking ASN.1 than this does, since dumpasn1 is a display program written
    to accept the widest possible range of input and not a compliance checker.
@@ -36,8 +38,16 @@
    that you use a utility like uudeview, which will strip virtually any kind
    of encoding (MIME, PEM, PGP, whatever) to recover the binary original.
 
-   You can use this code in whatever way you want, as long as you don't try to
-   claim you wrote it.
+   You can use this code in whatever way you want, as long as you don't try 
+   to claim you wrote it.
+
+   (Someone asked for clarification on what this means, treat it as a very 
+   mild form of the BSD license in which you're not required to include LONG 
+   LEGAL DISCLAIMERS IN ALL CAPS but just a small note in a corner somewhere 
+   (e.g. the back of a manual) that you're using the dumpasn1 code.  And if 
+   you do use it, please make sure you're using a recent version, I 
+   occasionally see screen shots from incredibly ancient versions that are
+   nowhere near as good as what current versions produce).
 
    Editing notes: Tabs to 4, phasers to malky (and in case anyone wants to
    complain about that, see "Program Indentation and Comprehensiblity",
@@ -56,7 +66,7 @@
 
 /* The update string, printed as part of the help screen */
 
-#define UPDATE_STRING	"8 June 2013"
+#define UPDATE_STRING	"8 August 2015"
 
 /* Useful defines */
 
@@ -241,6 +251,7 @@ typedef enum {
 	STR_PRINTABLE,			/* Check it's a PrintableString */
 	STR_IA5,				/* Check it's an IA5String */
 	STR_LATIN1,				/* Read and display string as latin-1 */
+	STR_UTF8,				/* Read and display string as UTF8 */
 	STR_BMP,				/* Read and display string as Unicode */
 	STR_BMP_REVERSED		/* STR_BMP with incorrect endianness */
 	} STR_OPTION;
@@ -340,9 +351,14 @@ static int fPos = 0;				/* Absolute position in data */
 
 static FILE *output;				/* Output stream */
 
-/* Information on an ASN.1 Object Identifier */
+/* OID data sizes.  Because of Microsoft's "encode random noise and call it 
+   an OID" approach, we maintain two size limits, a sane one and one capable
+   of holding the random-noise OID data, which we warn about */
 
-#define MAX_OID_SIZE	32
+#define MAX_OID_SIZE		40
+#define MAX_SANE_OID_SIZE	32
+
+/* Information on an ASN.1 Object Identifier */
 
 typedef struct tagOIDINFO {
 	struct tagOIDINFO *next;		/* Next item in list */
@@ -1206,7 +1222,7 @@ static void complain( const char *message, const int messageParam,
 	if( level < maxNestLevel )
 		{
 		if( !doPure )
-			fprintf( output, INDENT_STRING );
+			fprintf( output, "%s", INDENT_STRING );
 		doIndent( level + 1 );
 		}
 	fputs( "Error: ", output );
@@ -1224,11 +1240,11 @@ static void complainLength( const ASN1_ITEM *item, const int level )
 	if( level < maxNestLevel )
 		{
 		if( !doPure )
-			fprintf( output, INDENT_STRING );
+			fprintf( output, "%s", INDENT_STRING );
 		doIndent( level + 1 );
 		}
 #endif /* 0 */
-	fprintf( output, "Error: %s has invalid length %d.\n", 
+	fprintf( output, "Error: %s has invalid length %ld.\n", 
 			 idstr( item->tag ), item->length );
 	noErrors++;
 	}
@@ -1244,7 +1260,7 @@ static void complainLengthCanonical( const ASN1_ITEM *item, const int level )
 	if( level < maxNestLevel )
 		{
 		if( !doPure )
-			fprintf( output, INDENT_STRING );
+			fprintf( output, "%s", INDENT_STRING );
 		doIndent( level + 1 );
 		}
 #endif /* 0 */
@@ -1264,7 +1280,7 @@ static void complainInt( const BYTE *intValue, const int level )
 	if( level < maxNestLevel )
 		{
 		if( !doPure )
-			fprintf( output, INDENT_STRING );
+			fprintf( output, "%s", INDENT_STRING );
 		doIndent( level + 1 );
 		}
 	fprintf( output, "Error: Integer '%02X %02X ...' has non-DER encoding.\n", 
@@ -1319,8 +1335,9 @@ static int displayUnicode( const wchar_t wCh, const int level )
 						
 		/* To output Unicode to the Win32 console we need to switch the 
 		   output stream to Unicode-16 mode, but the following may also 
-		   depend on which code page is currently set for the console and 
-		   which font is being used */
+		   depend on which code page is currently set for the console, which 
+		   font is being used, and the phase of the moon (including the moons
+		   for Mars and Jupiter) */
 		fflush( output );
 		oldmode = _setmode( fileno( output ), _O_U16TEXT );
 		fputwc( wCh, output );
@@ -1412,12 +1429,15 @@ static void printValue( FILE *inFile, const int valueLength,
 		}
 	fPos += valueLength;
 
-	/* Display the integer value and any associated warnings */
+	/* Display the integer value and any associated warnings.  Note that 
+	   this will display an incorrectly-encoded integer as a negative value
+	   rather than the unsigned value that was probably intended to 
+	   emphasise that it's incorrect */
 	printString( level, " %ld\n", value );
 	if( warnNonDER )
 		complainInt( intBuffer, level );
 	if( warnNegative )
-		complain( "Integer has a negative value", 0, level );
+		complain( "Integer is encoded as a negative value", 0, level );
 	}
 
 /* Dump data as a string of hex digits up to a maximum of 128 bytes */
@@ -1479,6 +1499,12 @@ static void dumpHex( FILE *inFile, long length, int level,
 				}
 			}
 		ch = getc( inFile );
+		if( ch == EOF )
+			{
+			printString( level, "%c", '\n' );
+			complain( "Unexpected EOF, %d bytes missing", length - i, level );
+			return;
+			}
 		printString( level, "%s%02X", ( i % lineLength ) ? " " : "", ch );
 		printable[ i % 8 ] = ( ch >= ' ' && ch < 127 ) ? ch : '.';
 		fPos++;
@@ -1541,7 +1567,7 @@ static void dumpHex( FILE *inFile, long length, int level,
 		if( warnPadding )
 			complainInt( intBuffer, level );
 		if( warnNegative )
-			complain( "Integer has a negative value", 0, level );
+			complain( "Integer is encoded as a negative value", 0, level );
 		}
 	}
 
@@ -1659,11 +1685,11 @@ static int oidToString( char *textOID, int *textOIDlength,
 					}
 				length = sprintf( textOID, "%ld %ld", x, y );
 
-				/* An insane ITU facility lets people register UUIDs as OIDs
-				   (see http://www.itu.int/ITU-T/asn1/uuid.html), if we find
-				   one of these, which live under the arc '2 25' = 0x69 we
-				   have to continue decoding the OID as a UUID instead of a
-				   standard OID */
+				/* A totally stupid ITU facility lets people register UUIDs 
+				   as OIDs (see http://www.itu.int/ITU-T/asn1/uuid.html), if 
+				   we find one of these, which live under the arc '2 25' = 
+				   0x69 we have to continue decoding the OID as a UUID 
+				   instead of a standard OID */
 				if( data == 0x69 )
 					isUUID = TRUE;
 				}
@@ -1703,7 +1729,10 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
 		{
 		bitString = fgetc( inFile );
 		if( bitString == EOF )
+			{
+			noBits = 0;
 			errorStr = "Truncated BIT STRING data";
+			}
 		fPos++;
 		}
 	for( i = noBits - 8; i > 0; i -= 8 )
@@ -1858,6 +1887,40 @@ static void displayString( FILE *inFile, long length, int level,
 				ungetc( wCh & 0xFF, inFile );
 				}
 			}
+		if( strOption == STR_UTF8 && ( ch & 0x80 ) )
+			{
+			const int secondCh = getc( inFile );
+			wchar_t wCh;
+
+			/* It's a multibyte UTF8 character, read it as a widechar */
+			if( ( ch & 0xE0 ) == 0xC0 )		/* 111xxxxx -> 110xxxxx */
+				{
+				/* 2-byte character in the range 0x80...0x7FF */
+				wCh = ( ( ch & 0x1F ) << 6 ) | ( secondCh & 0x3F );
+				i++;		/* We've read 2 characters */
+				fPos += 2;
+				}
+			else
+				{
+				if( ( ch & 0xF0 ) == 0xE0 )	/* 1111xxxx -> 1110xxxx */
+					{
+					const int thirdCh = getc( inFile );
+
+					/* 3-byte character in the range 0x800...0xFFFF */
+					wCh = ( ( ch & 0x1F ) << 12 ) | \
+						  ( ( secondCh & 0x3F ) << 6 ) | \
+						  ( thirdCh & 0x3F );
+					}
+				else
+					wCh = '.';
+				i += 2;		/* We've read 3 characters */
+				fPos += 3;
+				}
+			if( !displayUnicode( wCh, level ) )
+				printString( level, "%c", '.' );
+			lineLength++;
+			continue;
+			}
 #endif /* __WIN32__ || __UNIX__ || __OS390__ */
 		switch( strOption )
 			{
@@ -2019,6 +2082,8 @@ static int getItem( FILE *inFile, ASN1_ITEM *item )
 		do
 			{
 			value = fgetc( inFile );
+			if( value == EOF )
+				return( FALSE );
 			tag = ( tag << 7 ) | ( value & 0x7F );
 			item->header[ index++ ] = value;
 			fPos++;
@@ -2073,7 +2138,7 @@ static int getItem( FILE *inFile, ASN1_ITEM *item )
 
 		/* Check for the length being less then 128, which means it 
 		   shouldn't be encoded as a long length */
-		if( item->length < 128 )
+		if( !item->indefinite && item->length < 128 )
 			item->nonCanonical = lengthStart;
 
 		/* Check for the first 9 bits of the length being identical and
@@ -2251,6 +2316,8 @@ static STR_OPTION checkForText( FILE *inFile, const int length )
 		   short strings are used in some places (eg PKCS #12 files) as
 		   IDs */
 		sampleLength = fread( buffer, 1, sampleLength, inFile );
+		if( sampleLength <= 0 )
+			return( STR_NONE );
 		fseek( inFile, -sampleLength, SEEK_CUR );
 		for( i = 0; i < sampleLength; i++ )
 			{
@@ -2264,6 +2331,8 @@ static STR_OPTION checkForText( FILE *inFile, const int length )
 
 	/* Check for ASCII-looking text */
 	sampleLength = fread( buffer, 1, sampleLength, inFile );
+	if( sampleLength <= 0 )
+		return( STR_NONE );
 	fseek( inFile, -sampleLength, SEEK_CUR );
 	if( isdigit( byteToInt( buffer[ 0 ] ) ) && \
 		( length == 13 || length == 15 ) && \
@@ -2371,12 +2440,15 @@ static void dumpHeader( FILE *inFile, const ASN1_ITEM *item, const int level )
 
 		for( i = 0; i < extraLen; i++ )
 			{
-			int ch = fgetc( inFile );
+			const int ch = fgetc( inFile );
 
-			if( feof( inFile ) )
-				extraLen = i;	/* Exit loop and get fseek() correct */
-			else
-				printString( level, " %02X", ch );
+			if( ch == EOF )
+				{
+				/* Exit loop and get fseek() offset correct */
+				extraLen = i;
+				break;
+				}
+			printString( level, " %02X", ch );
 			}
 		fseek( inFile, -extraLen, SEEK_CUR );
 		}
@@ -2653,17 +2725,27 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			int length, isValid;
 
 			/* Hierarchical Object Identifier */
-			if( item->length < 3 )	
-				complainLength( item, level );
-			if( item->length > MAX_OID_SIZE )
+			if( item->length <= 0 || item->length >= MAX_OID_SIZE )
 				{
 				fflush( stdout );
 				fprintf( stderr, "\nError: Object identifier length %ld too "
 						 "large.\n", item->length );
 				exit( EXIT_FAILURE );
 				}
-			fread( buffer, 1, ( size_t ) item->length, inFile );
+			length = fread( buffer, 1, ( size_t ) item->length, inFile );
 			fPos += item->length;
+			if( item->length < 3 )	
+				{
+				fputs( ".\n", output );
+				complainLength( item, level );
+				break;
+				}
+			if( length < item->length )
+				{
+				fputs( ".\n", output );
+				complain( "Invalid OID data", 0, level );
+				break;
+				}
 			if( ( oidInfo = getOIDinfo( buffer, ( int ) item->length ) ) != NULL )
 				{
 				/* Convert the binary OID to text form */
@@ -2709,6 +2791,13 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			isValid = oidToString( textOID, &length, buffer, 
 								   ( int ) item->length );
 			printString( level, " '%s'\n", textOID );
+			if( item->length > MAX_SANE_OID_SIZE )
+				{
+				/* This only occurs with Microsoft's "encode random noise 
+				   and call it an OID" values, so we warn about the fact 
+				   that it's not really an OID */
+				complain( "OID contains random garbage", 0, level );
+				}
 			if( !isValid )
 				complain( "OID has invalid encoding", 0, level );
 			if( item->nonCanonical )
@@ -2730,13 +2819,13 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 		case UNIVERSALSTRING:
 		case NUMERICSTRING:
 		case VIDEOTEXSTRING:
-		case UTF8STRING:
-			displayString( inFile, item->length, level, STR_NONE );
+		case PRINTABLESTRING:
+			displayString( inFile, item->length, level, STR_PRINTABLE );
 			if( item->nonCanonical )
 				complainLengthCanonical( item, level );
 			break;
-		case PRINTABLESTRING:
-			displayString( inFile, item->length, level, STR_PRINTABLE );
+		case UTF8STRING:
+			displayString( inFile, item->length, level, STR_UTF8 );
 			if( item->nonCanonical )
 				complainLengthCanonical( item, level );
 			break;
@@ -2916,6 +3005,9 @@ static int printAsn1( FILE *inFile, const int level, long length,
 					{
 					const int ch = fgetc( inFile );
 
+					if( ch == EOF )
+						return( 0 );
+
 					/* No object can be one byte long, try and recover.  This
 					   only works sometimes because it can be caused by
 					   spurious data in an OCTET STRING hole or an incorrect
@@ -3146,11 +3238,11 @@ int main( int argc, char *argv[] )
 					   garbage all over it for files larger than about 16K
 					   (which isn't), so we have to make sure that the
 					   stdout handle is pointed to something somewhere */
-					freopen( "nul", "w", stdout );
+					( void ) freopen( "nul", "w", stdout );
 #elif defined( __UNIX__ )
 					/* Safety feature in case any Unix libc is as broken
 					   as the Win32 version */
-					freopen( "/dev/null", "w", stdout );
+					( void ) freopen( "/dev/null", "w", stdout );
 #else
 					fclose( stdout );
 #endif /* OS-specific bypassing of stdout */
